@@ -6,6 +6,7 @@
 #include <vector>
 #include <valarray>
 #include <map>
+#include <forward_list>
 #include <cctype>
 #include <cmath>
 
@@ -15,6 +16,10 @@
 
 #include "DataFrame.h" // has #include Common.h
 
+// forward declaration for SMap solver
+std::valarray < double > SVD( DataFrame    < double > A,
+                              std::valarray< double > B );
+
 // Enumerations
 enum class Method         { None, Embed, Simplex, SMap, CCM };
 enum class DistanceMetric { Euclidean, Manhattan };
@@ -33,6 +38,20 @@ struct SMapValues {
     DataFrame< double > coefficients;
 };
 
+// Return object for CrossMap() worker function
+struct CrossMapValues {
+    DataFrame< double > LibStats;     // mean libsize, rho, RMSE, MAE
+    DataFrame< double > PredictStats; // each predict libsize, rho, RMSE, MAE
+    std::forward_list< DataFrame< double > > Predictions;
+};
+
+// Return object for CCM() with two CrossMapValues
+struct CCMValues {
+    DataFrame< double > AllLibStats;  // unified mean libsize, rho, RMSE, MAE
+    CrossMapValues CrossMap1;
+    CrossMapValues CrossMap2;
+};
+
 struct MultiviewValues {
     DataFrame< double > Combo_rho;              // col_i..., rho, MAE, RMSE
     DataFrame< double > Predictions;
@@ -77,6 +96,9 @@ std::string increment_datetime_str( std::string datetime1,
                                     std::string datetime2,
                                     int         tp );
 
+bool DistanceCompare( const std::pair<double, size_t> &x,
+                      const std::pair<double, size_t> &y );
+
 // API functions Embed() and MakeBlock() are in Embed.h Embed.cc
 
 DataFrame<double> Simplex( std::string pathIn          = "./data/",
@@ -88,7 +110,7 @@ DataFrame<double> Simplex( std::string pathIn          = "./data/",
                            int         E               = 0,
                            int         Tp              = 1,
                            int         knn             = 0,
-                           int         tau             = 1,
+                           int         tau             = -1,
                            int         exclusionRadius = 0,
                            std::string colNames        = "",
                            std::string targetName      = "",
@@ -104,7 +126,7 @@ DataFrame<double> Simplex( DataFrame< double > &dataFrameIn,
                            int         E               = 0,
                            int         Tp              = 1,
                            int         knn             = 0,
-                           int         tau             = 1,
+                           int         tau             = -1,
                            int         exclusionRadius = 0,
                            std::string colNames        = "",
                            std::string targetName      = "",
@@ -112,6 +134,10 @@ DataFrame<double> Simplex( DataFrame< double > &dataFrameIn,
                            bool        const_predict   = false,
                            bool        verbose         = true );
 
+// SMap is a special case since it can be called with a function pointer
+// to the SVD solver. This is done so that interfaces such as pybind11
+// can provide their own object for the solver.
+// 1) Data path/file with default SVD (LAPACK) assigned in Smap.cc 2)
 SMapValues SMap( std::string pathIn          = "./data/",
                  std::string dataFile        = "",
                  std::string pathOut         = "./",
@@ -121,7 +147,7 @@ SMapValues SMap( std::string pathIn          = "./data/",
                  int         E               = 0,
                  int         Tp              = 1,
                  int         knn             = 0,
-                 int         tau             = 1,
+                 int         tau             = -1,
                  double      theta           = 0,
                  int         exclusionRadius = 0,
                  std::string columns         = "",
@@ -132,6 +158,7 @@ SMapValues SMap( std::string pathIn          = "./data/",
                  bool        const_predict   = false,
                  bool        verbose         = true );
 
+// 2) DataFrame with default SVD (LAPACK) assigned in Smap.cc 2)
 SMapValues SMap( DataFrame< double > &dataFrameIn,
                  std::string pathOut         = "./",
                  std::string predictFile     = "",
@@ -140,7 +167,7 @@ SMapValues SMap( DataFrame< double > &dataFrameIn,
                  int         E               = 0,
                  int         Tp              = 1,
                  int         knn             = 0,
-                 int         tau             = 1,
+                 int         tau             = -1,
                  double      theta           = 0,
                  int         exclusionRadius = 0,
                  std::string columns         = "",
@@ -151,38 +178,85 @@ SMapValues SMap( DataFrame< double > &dataFrameIn,
                  bool        const_predict   = false,
                  bool        verbose         = true );
 
-DataFrame<double> CCM( std::string pathIn       = "./data/",
-                       std::string dataFile     = "",
-                       std::string pathOut      = "./",
-                       std::string predictFile  = "",
-                       int         E            = 0,
-                       int         Tp           = 0,
-                       int         knn          = 0,
-                       int         tau          = 1,
-                       std::string colNames     = "",
-                       std::string targetName   = "",
-                       std::string libSizes_str = "",
-                       int         sample       = 0,
-                       bool        random       = true,
-                       bool        replacement  = false,
-                       unsigned    seed         = 0,     // seed=0: use RNG
-                       bool        verbose      = true );
+// 3) Data path/file with external solver object, init to default SVD
+SMapValues SMap( std::string pathIn          = "./data/",
+                 std::string dataFile        = "",
+                 std::string pathOut         = "./",
+                 std::string predictFile     = "",
+                 std::string lib             = "",
+                 std::string pred            = "",
+                 int         E               = 0,
+                 int         Tp              = 1,
+                 int         knn             = 0,
+                 int         tau             = -1,
+                 double      theta           = 0,
+                 int         exclusionRadius = 0,
+                 std::string columns         = "",
+                 std::string target          = "",
+                 std::string smapFile        = "",
+                 std::string derivatives     = "",
+                 std::valarray<double> (*solver)(DataFrame < double >,
+                                             std::valarray < double >) = &SVD,
+                 bool        embedded        = false,
+                 bool        const_predict   = false,
+                 bool        verbose         = true );
 
-DataFrame<double> CCM( DataFrame< double > dataFrameIn,
-                       std::string         pathOut      = "./",
-                       std::string         predictFile  = "",
-                       int                 E            = 0,
-                       int                 Tp           = 0,
-                       int                 knn          = 0,
-                       int                 tau          = 1,
-                       std::string         colNames     = "",
-                       std::string         targetName   = "",
-                       std::string         libSizes_str = "",
-                       int                 sample       = 0,
-                       bool                random       = true,
-                       bool                replacement  = false,
-                       unsigned            seed         = 0, // seed=0: use RNG
-                       bool                verbose      = true );
+// 4) DataFrame with external solver object, init to default SVD
+SMapValues SMap( DataFrame< double > &dataFrameIn,
+                 std::string pathOut         = "./",
+                 std::string predictFile     = "",
+                 std::string lib             = "",
+                 std::string pred            = "",
+                 int         E               = 0,
+                 int         Tp              = 1,
+                 int         knn             = 0,
+                 int         tau             = -1,
+                 double      theta           = 0,
+                 int         exclusionRadius = 0,
+                 std::string columns         = "",
+                 std::string target          = "",
+                 std::string smapFile        = "",
+                 std::string derivatives     = "",
+                 std::valarray<double> (*solver)(DataFrame < double >,
+                                             std::valarray < double >) = &SVD,
+                 bool        embedded        = false,
+                 bool        const_predict   = false,
+                 bool        verbose         = true );
+
+CCMValues CCM( std::string pathIn       = "./data/",
+               std::string dataFile     = "",
+               std::string pathOut      = "./",
+               std::string predictFile  = "",
+               int         E            = 0,
+               int         Tp           = 0,
+               int         knn          = 0,
+               int         tau          = -1,
+               std::string colNames     = "",
+               std::string targetName   = "",
+               std::string libSizes_str = "",
+               int         sample       = 0,
+               bool        random       = true,
+               bool        replacement  = false,
+               unsigned    seed         = 0,     // seed=0: use RNG
+               bool        includeData  = false,
+               bool        verbose      = true );
+
+CCMValues CCM( DataFrame< double > dataFrameIn,
+               std::string         pathOut      = "./",
+               std::string         predictFile  = "",
+               int                 E            = 0,
+               int                 Tp           = 0,
+               int                 knn          = 0,
+               int                 tau          = -1,
+               std::string         colNames     = "",
+               std::string         targetName   = "",
+               std::string         libSizes_str = "",
+               int                 sample       = 0,
+               bool                random       = true,
+               bool                replacement  = false,
+               unsigned            seed         = 0, // seed=0: use RNG
+               bool                includeData  = false,
+               bool                verbose      = true );
 
 MultiviewValues Multiview( std::string pathIn          = "./",
                            std::string dataFile        = "",
@@ -190,14 +264,16 @@ MultiviewValues Multiview( std::string pathIn          = "./",
                            std::string predictFile     = "",
                            std::string lib             = "",
                            std::string pred            = "",
-                           int         E               = 0,
+                           int         D               = 0,
+                           int         E               = 1,
                            int         Tp              = 1,
                            int         knn             = 0,
-                           int         tau             = 1,
+                           int         tau             = -1,
                            std::string columns         = "",
                            std::string target          = "",
                            int         multiview       = 0,
                            int         exclusionRadius = 0,
+                           bool        trainLib        = true,
                            bool        verbose         = false,
                            unsigned    nThreads        = 4 );
 
@@ -206,14 +282,16 @@ MultiviewValues Multiview( DataFrame< double >,
                            std::string predictFile     = "",
                            std::string lib             = "",
                            std::string pred            = "",
-                           int         E               = 0,
+                           int         D               = 0,
+                           int         E               = 1,
                            int         Tp              = 1,
                            int         knn             = 0,
-                           int         tau             = 1,
+                           int         tau             = -1,
                            std::string columns         = "",
                            std::string target          = "",
                            int         multiview       = 0,
                            int         exclusionRadius = 0,
+                           bool        trainLib        = true,
                            bool        verbose         = false,
                            unsigned    nThreads        = 4 );
 
@@ -225,7 +303,7 @@ DataFrame<double> EmbedDimension( std::string pathIn      = "./data/",
                                   std::string pred        = "",
                                   int         maxE        = 10,
                                   int         Tp          = 1,
-                                  int         tau         = 1,
+                                  int         tau         = -1,
                                   std::string colNames    = "",
                                   std::string targetName  = "",
                                   bool        embedded    = false,
@@ -239,7 +317,7 @@ DataFrame<double> EmbedDimension( DataFrame< double > &dataFrameIn,
                                   std::string pred        = "",
                                   int         maxE        = 10,
                                   int         Tp          = 1,
-                                  int         tau         = 1,
+                                  int         tau         = -1,
                                   std::string colNames    = "",
                                   std::string targetName  = "",
                                   bool        embedded    = false,
@@ -254,7 +332,7 @@ DataFrame<double> PredictInterval( std::string pathIn      = "./data/",
                                    std::string pred        = "",
                                    int         maxTp       = 10,
                                    int         E           = 0,
-                                   int         tau         = 1,
+                                   int         tau         = -1,
                                    std::string colNames    = "",
                                    std::string targetName  = "",
                                    bool        embedded    = false,
@@ -268,7 +346,7 @@ DataFrame<double> PredictInterval( DataFrame< double > &dataFrameIn,
                                    std::string pred        = "",
                                    int         maxTp       = 10,
                                    int         E           = 0,
-                                   int         tau         = 1,
+                                   int         tau         = -1,
                                    std::string colNames    = "",
                                    std::string targetName  = "",
                                    bool        embedded    = false,
@@ -285,7 +363,7 @@ DataFrame<double> PredictNonlinear( std::string pathIn      = "./data/",
                                     int         E           = 0,
                                     int         Tp          = 1,
                                     int         knn         = 0,
-                                    int         tau         = 1,
+                                    int         tau         = -1,
                                     std::string colNames    = "",
                                     std::string targetName  = "",
                                     bool        embedded    = false,
@@ -301,7 +379,7 @@ DataFrame<double> PredictNonlinear( DataFrame< double > &dataFrameIn,
                                     int         E           = 0,
                                     int         Tp          = 1,
                                     int         knn         = 0,
-                                    int         tau         = 1,
+                                    int         tau         = -1,
                                     std::string colNames    = "",
                                     std::string targetName  = "",
                                     bool        embedded    = false,
